/*
 * Copyright 2016 Capital One Services, LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __GIRAFFEZ_BUFFER_H
#define __GIRAFFEZ_BUFFER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "common.h"


typedef struct buffer_t {
    size_t length;
    size_t pos;
    char   *data;
} buffer_t;

buffer_t* buffer_new(int buffer_size);
void      buffer_write(buffer_t *b, char *data, int length);
void      buffer_reset(buffer_t *b, size_t n);
void      buffer_writef(buffer_t *b, const char *fmt, ...);


#ifdef __cplusplus
}
#endif

#endif
